// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//


package de.uka.ilkd.key.strategy.feature;

import de.uka.ilkd.key.logic.PosInOccurrence;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.ldt.IntegerLDT;
import de.uka.ilkd.key.logic.op.Function;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.rule.TacletApp;
import de.uka.ilkd.key.strategy.LongRuleAppCost;
import de.uka.ilkd.key.strategy.TopRuleAppCost;
import de.uka.ilkd.key.strategy.termProjection.ProjectionToTerm;
import de.uka.ilkd.key.strategy.termfeature.BinarySumTermFeature;
import de.uka.ilkd.key.strategy.termfeature.ConstTermFeature;
import de.uka.ilkd.key.strategy.termfeature.OperatorTF;
import de.uka.ilkd.key.strategy.termfeature.SubTermFeature;
import de.uka.ilkd.key.strategy.termfeature.TermFeature;


/**
 * Feature that returns zero iff each monomial of one polynomial is smaller than
 * all monomials of a second polynomial
 */
public class MonomialsSmallerThanFeature extends SmallerThanFeature {

    private final TermFeature  hasCoeff;
    private final IntegerLDT numbers;

    private final ProjectionToTerm left, right;
    private final Function add, mul, Z;


    private MonomialsSmallerThanFeature(ProjectionToTerm left, ProjectionToTerm right, 
                                        IntegerLDT numbers) {
        this.left = left;
        this.right = right;
        this.numbers = numbers;
        this.add = numbers.getArithAddition ();
        this.mul = numbers.getArithMultiplication ();
        this.Z = numbers.getNumberSymbol ();
        
        hasCoeff = createHasCoeffTermFeature ( numbers );
    }
    
    static TermFeature createHasCoeffTermFeature(final IntegerLDT numbers) {
        return
            BinarySumTermFeature.createSum (
                  OperatorTF.create ( numbers.getArithMultiplication() ),
                  SubTermFeature.create ( new TermFeature[] {
                        ConstTermFeature.createConst ( LongRuleAppCost.ZERO_COST ),
                        OperatorTF.create ( numbers.getNumberSymbol()) } ) );
    }

    public static Feature create(ProjectionToTerm left, ProjectionToTerm right, 
                                 IntegerLDT numbers) {
        return new MonomialsSmallerThanFeature ( left, right, numbers );
    }
    
    protected boolean filter(TacletApp app, PosInOccurrence pos, Goal goal) {
        final MonomialCollector m1 = new MonomialCollector ();
        m1.collect ( left.toTerm ( app, pos, goal ) );
        final MonomialCollector m2 = new MonomialCollector ();
        m2.collect ( right.toTerm ( app, pos, goal ) );

        return lessThan ( m1.getResult(), m2.getResult() );
    }

    /**
     * this overwrites the method of <code>SmallerThanFeature</code>
     */
    protected boolean lessThan(Term t1, Term t2) {

        // here, the ordering is graded concerning multiplication on integers
        final int t1Deg = degree ( t1 );
        final int t2Deg = degree ( t2 );
        if ( t1Deg < t2Deg ) return true;
        if ( t1Deg > t2Deg ) return false;
        
        return super.lessThan ( t1, t2 );
    }

    /**
     * @return the degree of a monomial (number of terms that are connected
     *         multiplicatively)
     */
    private int degree(Term t) {
        if ( t.op() == mul ) {
            return degree ( t.sub ( 0 ) ) + degree ( t.sub ( 1 ) );
        }
        return 1;
    }

    private class MonomialCollector extends Collector {
        protected void collect(Term te) {
            if ( te.op () == add ) {
                collect ( te.sub ( 0 ) );
                collect ( te.sub ( 1 ) );
            } else if ( te.op() == Z ) {
              // nothing  
            } else {
                addTerm ( stripOffLiteral ( te ) );
            }
        }

        private Term stripOffLiteral(Term te) {
            if ( ! ( hasCoeff.compute ( te ) instanceof TopRuleAppCost ) )
                // we leave out literals/coefficients on the right, because we
                // do not want to compare these literals
                return te.sub ( 0 );
            return te;
        }
    }
}
